from datetime import timedelta
from importlib import reload
from unittest import skipIf

from bs4 import BeautifulSoup
from django.conf import settings
from django.template import Context, Template
from django.test import Client, TestCase
from django.urls import reverse
from requests.exceptions import ConnectionError

from dalec.proxy import ProxyPool
from dalec.tests_utils import DalecTestCaseMixin
from dalec_caldav import proxy

__all__ = ["DalecTests"]


class DalecTests(DalecTestCaseMixin, TestCase):
    def _cal_url(self, cal):
        if cal == "primary":
            cal = "7eb24728-09c0-3977-4fca-e7bec231f7a0"
        elif cal == "secondary":
            cal = "4fc45ea4-371e-81ab-ee54-295eb8ec28a6"
        return "%s/%s" % (settings.DALEC_CALDAV_BASE_URL, cal)

    def test_caldav_with_wrong_url(self):
        wrong_url = "https://doesnoexist.dalec.webu.coop/"
        with self.settings(DALEC_CALDAV_BASE_URL=wrong_url):
            ProxyPool.unregister("caldav")
            reload(proxy)
            dalec_caldav = ProxyPool.get("caldav")
            with self.assertRaises(ConnectionError):
                dalec_caldav.refresh(
                    content_type="event",
                    channel="url",
                    channel_object=wrong_url,
                    force=True,
                )
        ProxyPool.unregister("caldav")
        reload(proxy)

    def test_retrieve_events_from_multiple_calendars(self):
        kwargs = {"app": "caldav", "content_type": "event", "channel": "url"}
        url = reverse("dalec_fetch_content", kwargs=kwargs)
        client = Client()
        response = client.post(
            url,
            # no channel objects: lets fetch events from all calendars
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        qs = self.content_model.objects.filter(
            app="caldav",
            content_type="event",
            channel="url",
        )
        self.assertEqual(qs.count(), 6)
        contents = dict((c.content_id.split("-", 1)[0], c) for c in qs)
        self.assertIn("88bb03ce", contents)
        self.assertEqual(contents["88bb03ce"].content_data["calendar_displayname"], "Secondary")
        self.assertIn("0239cc8f", contents)
        self.assertEqual(contents["0239cc8f"].content_data["calendar_displayname"], "Primary")
        self.assertIn("04bbb7da", contents)
        self.assertIn("f54e19f1", contents)
        self.assertIn("b843d43a", contents)
        self.assertIn("8c6c5b2d", contents)

        # check events are returned in default order (last_updated first)
        orig_keys = [k for k in contents.keys()]
        expect_keys = [
            c.content_id.split("-", 1)[0]
            for c in sorted(contents.values(), key=lambda c: c.last_update_dt, reverse=True)
        ]
        self.assertEqual(orig_keys, expect_keys)

    def test_retrieve_events_from_unique_calendars(self):
        kwargs = {"app": "caldav", "content_type": "event", "channel": "url"}
        url = reverse("dalec_fetch_content", kwargs=kwargs)
        client = Client()
        response = client.post(
            url,
            '{"channelObjects": ["%s"]}' % self._cal_url("secondary"),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        qs = self.content_model.objects.filter(
            app="caldav",
            content_type="event",
            channel="url",
        )
        self.assertEqual(qs.count(), 1)
        contents = dict((c.content_id.split("-", 1)[0], c) for c in qs)
        self.assertIn("88bb03ce", contents)
        self.assertEqual(contents["88bb03ce"].content_data["calendar_displayname"], "Secondary")

    def test_dates(self):
        kwargs = {"app": "caldav", "content_type": "event", "channel": "url"}
        url = reverse("dalec_fetch_content", kwargs=kwargs)
        client = Client()
        response = client.post(
            url,
            '{"channelObjects": ["%s"]}' % self._cal_url("primary"),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)
        qs = self.content_model.objects.filter(
            app="caldav",
            content_type="event",
            channel="url",
            channel_object=self._cal_url("primary"),
        )
        self.assertEqual(qs.count(), 5)
        contents = dict((c.content_id.split("-", 1)[0], c) for c in qs)
        # event on 2 days, without hours
        self.assertEqual(contents["0239cc8f"].content_data["duration"]["days"], 2)
        self.assertIsNone(contents["0239cc8f"].content_data["start_time"])
        self.assertIsNone(contents["0239cc8f"].content_data["end_time"])
        self.assertEqual(contents["0239cc8f"].content_data["start_date"], "2035-07-04")
        self.assertEqual(contents["0239cc8f"].content_data["end_date"], "2035-07-06")

        # event on 1 day, without hours
        self.assertEqual(contents["04bbb7da"].content_data["duration"]["days"], 1)
        self.assertIsNone(contents["04bbb7da"].content_data["start_time"])
        self.assertIsNone(contents["04bbb7da"].content_data["end_time"])
        self.assertEqual(contents["04bbb7da"].content_data["start_date"], "2035-07-02")
        self.assertEqual(contents["04bbb7da"].content_data["end_date"], "2035-07-03")

        # event on 9+ days, with hours
        self.assertEqual(contents["f54e19f1"].content_data["duration"]["days"], 9)
        self.assertEqual(contents["f54e19f1"].content_data["duration"]["seconds"], 14400)
        self.assertEqual(contents["f54e19f1"].content_data["duration"]["total_seconds"], 792000)
        self.assertEqual(contents["f54e19f1"].content_data["start_time"], "20:10:00")
        self.assertEqual(contents["f54e19f1"].content_data["end_time"], "00:10:00")
        self.assertEqual(contents["f54e19f1"].content_data["start_date"], "2030-10-10")
        self.assertEqual(contents["f54e19f1"].content_data["end_date"], "2030-10-20")

        # event for some hours
        self.assertEqual(contents["8c6c5b2d"].content_data["duration"]["days"], 0)
        self.assertEqual(contents["8c6c5b2d"].content_data["duration"]["seconds"], 2520)
        self.assertEqual(contents["8c6c5b2d"].content_data["duration"]["total_seconds"], 2520)
        self.assertEqual(contents["8c6c5b2d"].content_data["start_time"], "00:42:00")
        self.assertEqual(contents["8c6c5b2d"].content_data["end_time"], "01:24:00")
        self.assertEqual(contents["8c6c5b2d"].content_data["start_date"], "2042-02-11")
        self.assertEqual(contents["8c6c5b2d"].content_data["end_date"], "2042-02-11")

    def test_dates_in_template(self):
        # load contents
        kwargs = {
            "app": "caldav",
            "content_type": "event",
            "channel": "url",
            "ordered_by": "dtstart",
        }
        url = reverse("dalec_fetch_content", kwargs=kwargs)
        client = Client()
        response = client.post(
            url,
            '{"channelObjects": ["%s"]}' % self._cal_url("primary"),
            content_type="application/json",
        )
        self.assertEqual(response.status_code, 200)

        t = Template(
            """
            {% load dalec %}
            {% dalec "caldav" "event" channel="url" channel_object=url ordered_by="dtstart" %}
        """
        )
        output = t.render(Context({"url": self._cal_url("primary")}))
        soup = BeautifulSoup(output, "html.parser")
        dalec_divs = soup.find_all(class_="dalec-item")
        date_divs = soup.find_all(class_="caldav-item-dates")
        self.assertEqual(len(date_divs), 5)

        times = dalec_divs[0].find_all("time")
        self.assertEqual(len(times), 2)
        self.assertEqual(times[0]["datetime"], "2030-10-10T20:10:00+02:00")
        self.assertEqual(times[1]["datetime"], "2030-10-20T00:10:00+02:00")

        times = dalec_divs[1].find_all("time")
        self.assertEqual(len(times), 1)
        self.assertEqual(times[0]["datetime"], "2035-07-02")

        times = dalec_divs[2].find_all("time")
        self.assertEqual(len(times), 2)
        self.assertEqual(times[0]["datetime"], "2035-07-04")
        self.assertEqual(times[1]["datetime"], "2035-07-06")

        times = dalec_divs[4].find_all("time")
        self.assertEqual(len(times), 2)
        self.assertEqual(times[0]["datetime"], "2042-02-11T00:42:00+01:00")
        self.assertEqual(times[1]["datetime"], "2042-02-11T01:24:00+01:00")

    @skipIf(
        bool(settings.DALEC_EXTRA_CALDAV_URLS) is False,
        "no extra caldavs to test. see README and local_settings.py if you want to",
    )
    def test_retrieve_events_from_other_caldavs(self):
        for name, conf in settings.DALEC_EXTRA_CALDAV_URLS.items():
            ProxyPool.unregister("caldav")
            new_settings = {
                "DALEC_CALDAV_BASE_URL": conf["url"],
                "DALEC_CALDAV_API_USERNAME": conf.get("username"),
                "DALEC_CALDAV_API_PASSWORD": conf.get("password"),
                "DALEC_CALDAV_SERCH_EVENT_END_TIMEDELTA": timedelta(days=365),
            }
            with self.settings(**new_settings):
                reload(proxy)
                dalec_caldav = ProxyPool.get("caldav")
            try:
                nb_created, nb_updated, nb_deleted = dalec_caldav.refresh(
                    content_type="event",
                    channel="url",
                    channel_object=conf["url"],
                    force=True,
                )
            except Exception as e:
                raise Exception("%s error: %s" % (name, e)) from e
            qs = self.content_model.objects.filter(
                app="caldav",
                content_type="event",
                channel="url",
                channel_object=conf["url"],
            )
            self.assertTrue(qs.exists())
            self.assertGreaterEqual(nb_created, 1)
            self.assertEqual(nb_updated, 0)
            self.assertGreaterEqual(nb_created, qs.count())
        ProxyPool.unregister("caldav")
        reload(proxy)



import os

import django

try:
    from .local_settings import *  # noqa: F403
except ImportError:
    pass

BASE_PATH = os.path.join(os.path.dirname(__file__), "..")
SECRET_KEY = "EX-TER-MI-NA-TE"

# If you set this to False, Django will not use timezone-aware datetimes.
USE_TZ = True
TIME_ZONE = "Europe/Paris"

INSTALLED_APPS = [
    "django.contrib.auth",
    "django.contrib.contenttypes",
    "django.contrib.sessions",
    "django.contrib.staticfiles",
    "tests",
    "dalec_prime",
    "dalec",
    "dalec_caldav",
]

if django.VERSION < (3, 2):
    INSTALLED_APPS.append("django_jsonfield_backport")

ROOT_URLCONF = "tests.urls"

TEMPLATES = [{"BACKEND": "django.template.backends.django.DjangoTemplates", "APP_DIRS": True}]

DATABASES = {
    "default": {
        "ENGINE": "django.db.backends.sqlite3",
        "NAME": os.path.join(BASE_PATH, "tests.sqlite3"),
    }
}

DALEC_CALDAV_BASE_URL = "http://localhost:5232/test/"
DALEC_CALDAV_API_USERNAME = "test"
DALEC_CALDAV_API_PASSWORD = "test"
# For tests, we want ALL events (except for DALEC_EXTRA_CALDAV_URLS: it will be 365 days max)
DALEC_CALDAV_SERCH_EVENT_END_TIMEDELTA = None

try:
    # should be defined in local_settings
    DALEC_EXTRA_CALDAV_URLS  # noqa: F405
except NameError:
    DALEC_EXTRA_CALDAV_URLS = {
        # should be with this format:
        # "nextclould-v27": {
        #     "url": "https://nextcloud.example.com/index.php/apps/calendar/p/xxx/CalendarName",
        #     "username": "test",  # optionnal, only if this caldav needs authentication
        #     "password": "test",  # optionnal, only if this caldav needs authentication
        # }
    }



# Future imports
from __future__ import annotations

# Standard libs
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from datetime import date
    from typing import Dict, Any, List
    from vobject.base import Component
    from caldav.objects import Calendar, CalendarObjectResource

from datetime import datetime, timedelta
from urllib.parse import urlparse

# Django imports
from django.conf import settings
from django.utils.timezone import make_aware, now


# DALEC imports
from caldav.davclient import DAVClient
from caldav.elements import dav
from caldav.objects import Event
from dalec import settings as app_settings
from dalec.proxy import Proxy

client = DAVClient(
    url=settings.DALEC_CALDAV_BASE_URL,
    username=settings.DALEC_CALDAV_API_USERNAME,
    password=settings.DALEC_CALDAV_API_PASSWORD,
)


class CaldavProxy(Proxy):
    """
    caldav dalec proxy to fetch the last event.
    """

    app = "caldav"

    def _fetch(
        self, nb: int, content_type: str, channel: str, channel_object: str
    ) -> Dict[str, dict]:
        if content_type == "event":
            return self._fetch_event(nb, channel, channel_object)
        raise ValueError(f"Invalid content_type {content_type}. Accepted: event.")

    def _fetch_event(self, nb: int, channel: str, channel_object: str) -> Dict[str, dict]:
        """
        Get latest events from calendar(s)
        """

        if channel == "url" and channel_object:
            if channel_object[-1] != "/":
                channel_object += "/"
            calendars = client.calendar(url=channel_object)
            if type(calendars) != list:
                calendars = [calendars]
        else:
            principal = client.principal()
            calendars = principal.calendars()

        contents = {}
        for calendar in calendars:
            calendar_infos = self._get_calendar_infos(calendar)
            for event in self._get_events(calendar):
                content = self._populate_content(event, calendar_infos)
                contents[content["id"]] = content
        return contents

    def _get_events(self, calendar: Calendar) -> List[CalendarObjectResource]:
        start_td = app_settings.get_setting(
            "CALDAV_SERCH_EVENT_START_TIMEDELTA", timedelta(days=-1)
        )
        end_td = app_settings.get_setting("CALDAV_SERCH_EVENT_END_TIMEDELTA", timedelta(days=365))
        search_kwargs = {
            "comp_class": Event,
            "start": now() + start_td if start_td is not None else None,
            "end": now() + end_td if end_td is not None else None,
        }
        return calendar.search(**search_kwargs)

    def _get_calendar_infos(self, calendar: Calendar) -> Dict[str, Any]:
        url = str(calendar.url)
        infos = {
            "type": "default",
            "display_name": calendar.get_property(dav.DisplayName()),
            "url": url,
        }
        if "/remote.php/dav/public-calendars/" in url:
            url_obj = urlparse(url)
            token = url_obj.path.split("/")[-2]
            nextcloud_calendar_url = "{}://{}/index.php/apps/calendar/p/{}".format(
                url_obj.scheme, url_obj.netloc, token
            )
            infos.update(
                {
                    "type": "nextcloud",
                    "nextcloud_calendar_url": nextcloud_calendar_url,
                    "token": token,
                }
            )
        return infos

    def _vobject_to_dict(self, vobject: Component) -> Dict[str, Any[str, int, datetime, date]]:
        content = {}
        for key, val in vobject.contents.items():
            if hasattr(val[0], "contents"):
                content[key] = self._vobject_to_dict(val[0])
            else:
                content[key] = val[0].value
        return content

    def _populate_content(self, event: Event, calendar_infos: Dict[str, Any]) -> dict:
        content = self._vobject_to_dict(event.vobject_instance.vevent)
        content["event_url"] = event.canonical_url
        content["dav_calendar_url"] = calendar_infos["url"]
        content["calendar_displayname"] = calendar_infos["display_name"]
        if "created" not in content:
            # some events accepted from an email in thunderbird can be without created info
            dates = []
            if "dtstamp" in content:
                dates.append(
                    make_aware(content["dtstamp"])
                    if not content["dtstamp"].tzinfo
                    else content["dtstamp"]
                )
            if "last-modified" in content:
                dates.append(
                    make_aware(content["last-modified"])
                    if not content["last-modified"].tzinfo
                    else content["last-modified"]
                )
            content["created"] = min(dates) if dates else None
        if "last-modified" not in content:
            content["last-modified"] = content["created"]
        if calendar_infos["type"] == "nextcloud":
            content["nextcloud_calendar_url"] = calendar_infos["nextcloud_calendar_url"]
        duration = event.get_duration()
        content.update(
            {
                "id": content["uid"],
                "creation_dt": (
                    make_aware(content["created"])
                    if not content["created"].tzinfo
                    else content["created"]
                ),
                "last_update_dt": (
                    make_aware(content["last-modified"])
                    if not content["last-modified"].tzinfo
                    else content["last-modified"]
                ),
                "duration": {
                    "days": duration.days,
                    "seconds": duration.seconds,
                    "total_seconds": int(duration.total_seconds()),
                },
            }
        )
        if isinstance(content["dtstart"], datetime):
            content["start_date"] = content["dtstart"].date()
            content["start_time"] = content["dtstart"].time()
        else:
            content["start_date"] = content["dtstart"]
            content["start_time"] = None
        if isinstance(content["dtend"], datetime):
            content["end_date"] = content["dtend"].date()
            content["end_time"] = content["dtend"].time()
        else:
            content["end_date"] = content["dtend"]
            content["end_time"] = None
        return content
